`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Adaptive Notch Filter
// 2020.12.12 Arata Kawamura
//////////////////////////////////////////////////////////////////////////////////

module axis_proc #(
    parameter SWITCH_WIDTH = 4, // Number of Slide Switch
    parameter DATA_WIDTH = 24
) (
    input wire clk,
    input wire [SWITCH_WIDTH-1:0] sw,
    
    //AXIS SLAVE INTERFACE
    input  wire [DATA_WIDTH-1:0] s_axis_data,
    input  wire s_axis_valid,
    output reg  s_axis_ready = 1'b1,
    input  wire s_axis_last,
    
    // AXIS MASTER INTERFACE
    output reg [DATA_WIDTH-1:0] m_axis_data = 1'b0,
    output reg m_axis_valid = 1'b0,
    input  wire m_axis_ready,
    output reg m_axis_last = 1'b0
);

    wire m_select      = m_axis_last;
    wire m_new_word    = (m_axis_valid == 1'b1 && m_axis_ready == 1'b1) ? 1'b1 : 1'b0;
    wire m_new_packet  = (m_new_word == 1'b1 && m_axis_last == 1'b1) ? 1'b1 : 1'b0;
    
    wire s_select      = s_axis_last;
    wire s_new_word    = (s_axis_valid == 1'b1 && s_axis_ready == 1'b1) ? 1'b1 : 1'b0;
    wire s_new_packet  = (s_new_word == 1'b1 && s_axis_last == 1'b1) ? 1'b1 : 1'b0;
    reg s_new_packet_r = 1'b0;
    

    //**********************************************************//
    //*        Setting for Fixed Point Calculation             *//
    //*                                                        *//
    //*        [integer(8bits)].[decimal(40bits)]              *//
    //*   ---> SignBit(8bits)_input(24bits)_0(16bits)          *//
    //*                                                        *//
    //**********************************************************//
    parameter TBL = 48;              // Total Bit Length including Sign Bit
    parameter DPL = 40;              // Decimal Part Length

    parameter M   = 1;               // Memory length for delayed signals

    //**********************************************************//
    //*                Declaration of Signals                   *//
    //**********************************************************//
    reg [TBL-1:0] xin  [1:0];        // input signal
    reg [TBL-1:0] yout [1:0];        // output signal
    reg [TBL-1:0] x_L  [0:M];        // left  delayed input  signal
    reg [TBL-1:0] x_R  [0:M];        // right delayed input  signal
    reg [TBL-1:0] y_L  [0:M];        // left  delayed output signal
    reg [TBL-1:0] y_R  [0:M];        // right delayed output signal
    integer   tn  = 0;               // input time index
    integer   tout= 0;               // output time index

    //**********************************************************//
    //*                Declaration of Paramters                *//
    //**********************************************************//
    reg signed [TBL-1:0] mu= (1)*(2**DPL)/10;    // mu = 0.1
    reg signed [TBL-1:0] p = (5)*(2**DPL)/10;    // p = 0.5
    reg signed [TBL-1:0] aL = 0;                 // aL = -(1+r)cos(w_N)
    reg signed [TBL-1:0] aR = 0;                 // aR = -(1+r)cos(w_N)
    reg signed [TBL-1:0] r = (8)*(2**DPL)/10;    // r = 0.8
    reg signed [TBL-1:0] UT =  (18)*(2**DPL)/10; // Upper Threshold = (1+r)  =  1.8
    reg signed [TBL-1:0] LT = (-18)*(2**DPL)/10; // Lower Threshold = -(1+r) = -1.8
    reg signed [TBL-1:0] uL [2:0];
    reg signed [TBL-1:0] uR [2:0];
    reg signed [TBL-1:0] y  [2:0];
    reg signed [TBL-1:0] e  [2:0];
    
    /* Examples of decalaration 
    reg signed [TBL-1:0] x;
    reg signed [TBL-1:0] u  =   (5)*(2**DPL);    // u = 5
    reg signed [TBL-1:0] h  =   (1)*(2**DPL)/10; // h = 0.1
    reg signed [TBL-1:0] a  =  (-1)*(2**DPL)/10; // a = -0.1
    reg signed [TBL-1:0] th =  (80)*(2**DPL);    // th = 80
    reg signed [TBL-1:0] b  = (-80)*(2**DPL);    // b = -80
    */
        
    always@(posedge clk) begin
        if (s_new_word == 1'b1)          // get 24bits input data and sign extend --> 48bits
            xin[s_select] <= { {8{s_axis_data[DATA_WIDTH-1]}}, s_axis_data, {16{1'b0}} };
        else if (s_new_packet_r == 1'b1) begin
        
            //**********************************************************//
            //*                                                        *//
            //*                  Signal Processing                     *//
            //*                                                        *//
            //*              Sampling Frequency 44.1kHz                *//
            //*                    2 channels                          *//
            //**********************************************************//
            x_L[tn]  = $signed( xin[0] );             // left  input signal (48bits)
            x_R[tn]  = $signed( xin[1] );             // right input signal (48bits)

            // non-blocking ŃtB^o͂Lq
            uL[0] = $signed( x_L[tn] ) - $signed( FPM( aL, uL[1] ) ) - $signed( FPM( r, uL[2] ) );
            uR[0] = $signed( x_R[tn] ) - $signed( FPM( aR, uR[1] ) ) - $signed( FPM( r, uR[2] ) );
            
            // Allpass filter output
            y[0]  = $signed( FPM( r, uL[0] ) ) + $signed( FPM( aL, uL[1] ) ) + $signed( uL[2] );
            y[1]  = $signed( FPM( r, uR[0] ) ) + $signed( FPM( aR, uR[1] ) ) + $signed( uR[2] );
            
            // Notch filter output
            e[0]  = $signed( x_L[tn] ) + $signed( y[0] );
            e[1]  = $signed( x_R[tn] ) + $signed( y[1] );
            
            // Set output signal
            y_L[0] = $signed( FPM(p, e[0]) );
            y_R[0] = $signed( FPM(p, e[1]) );
            
            // Adaptive Algorithm
            aL  <= $signed(aL) - $signed( FPM( FPM(mu, e[0]), uL[1]) );
            aR  <= $signed(aR) - $signed( FPM( FPM(mu, e[1]), uR[1]) );
            
            // Initialization of aL, aR when out of range
            if ( $signed(aL) > UT )
                aL <= 0;
            if ( $signed(aL) < LT )
                aL <= 0;
            if ( $signed(aR) > UT )
                aR <= 0;
            if ( $signed(aR) < LT )
                aR <= 0;

            // Shift
            uL[2] <= uL[1];
            uL[1] <= uL[0];
            uR[2] <= uR[1];
            uR[1] <= uR[0];

            tn      <= (tn  + 1)%M;                    // ͐M̎Ǘ
            tout    <= (tout+ 1)%M;                    // o͎=1O
            
            //**********************************************************//
            //*               End of Signal Processing                 *//
            //**********************************************************//
            //**************************************//
            //*                                    *//
            //*  sw[0]=0: Through                  *//
            //*  sw[0]=1: Processing               *//
            //*                                    *//
            //**************************************//
            yout[0] = (sw[0])? $signed(y_L[0]) : $signed(xin[0]); // left  channel output
            yout[1] = (sw[0])? $signed(y_R[0]) : $signed(xin[1]); // right channel output
        end
    end
        
    always@(posedge clk) begin
        s_new_packet_r <= s_new_packet;
    end
    
    always@(posedge clk)
        if (s_new_packet_r == 1'b1)
            m_axis_valid <= 1'b1;
        else if (m_new_packet == 1'b1)
            m_axis_valid <= 1'b0;
            
    always@(posedge clk)
        if (m_new_packet == 1'b1)
            m_axis_last <= 1'b0;
        else if (m_new_word == 1'b1)
            m_axis_last <= 1'b1;
            
    //************************************************************//
    //*  D/A <-- [24bits] <-- yout is [8bits].[24bits][16bits]   *//
    //***********************************************************s*//
    always@(m_axis_valid, yout[0], yout[1], m_select)
        if (m_axis_valid == 1'b1)
            m_axis_data = yout[m_select][39:16];
        else
            m_axis_data = 'b0;
            
    always@(posedge clk)
        if (s_new_packet == 1'b1)
            s_axis_ready <= 1'b0;
        else if (m_new_packet == 1'b1)
            s_axis_ready <= 1'b1;

    /////////////////////////////
    // Multiplication Function //
    /////////////////////////////
    function [TBL-1:0] FPM(input [TBL-1:0] IN1, input [TBL-1:0] IN2);
    reg signed [2*TBL-1:0] MUL;
    begin
        MUL = $signed(IN1) * $signed(IN2);
        MUL = MUL / (2**DPL);
        FPM = $signed({MUL[2*TBL-1], MUL[TBL-2:0]});
    end
    endfunction
    
    ///////////////////////
    // Division Function //
    ///////////////////////
    function [TBL-1:0] FPD(input [TBL-1:0] IN1, input [TBL-1:0] IN2);
    reg signed [2*TBL-1:0] DIV;
    begin
        DIV = $signed(IN1) * $signed(2**DPL);
        DIV = DIV / $signed(IN2);
        FPD = $signed({DIV[2*TBL-1], DIV[TBL-2:0]});
    end
    endfunction
            
endmodule
